/**
 * Copyright (c) 2009 - Lehigh University. Bethlehem, PA, USA.
 * All rights reserved.
 * This source code is a part of MARCHES Project. 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs, and the author attribution appear in all copies of this
 * software.
 *
 * IN NO EVENT SHALL LEHIGH UNIVERSITY BE LIABLE TO ANY PARTY FOR DIRECT, 
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LEHIGH
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *	
 * LEHIGH UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
 * AND LEHIGH UNIVERSITY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 **/

// modsDlg.cpp : implementation file
//

using System;
using System.Net;
using System.Diagnostics;
using System.Net.Sockets;
using System.Collections;
using System.Timers;
using System.Threading;

namespace cmpMassPing
{
    struct ICMPHeader
    {
        public const int ICMPHEADERLENGTH = 22;
        public byte type;
        public byte code;
        public ushort chksum;
        public ushort id;
        public ushort seq;
        public ulong timestamp;

        public byte[] toByteArray()
        {
            //If you know a better way to serialize this into a byte array, let me know
            int idx = 0;
            byte[] arResult = new byte[ICMPHEADERLENGTH];
            byte[] arTemp;

            arResult[0] = type;
            arResult[1] = code;
            idx += 2;

            arTemp = BitConverter.GetBytes(chksum);
            arTemp.CopyTo(arResult, idx);
            idx += arTemp.Length;
            
            arTemp = BitConverter.GetBytes(id);
            arTemp.CopyTo(arResult, idx);
            idx += arTemp.Length;
            
            arTemp = BitConverter.GetBytes(seq);
            arTemp.CopyTo(arResult, idx);
            idx += arTemp.Length;
            
            arTemp = BitConverter.GetBytes(timestamp);
            arTemp.CopyTo(arResult, idx);
            idx += arTemp.Length;
            
            return arResult;
        }

        static public ICMPHeader fromByteArray(byte[] arIn)
        {
            ICMPHeader icmph = new ICMPHeader();
            int idx = 0;
            byte[] arTemp;

            icmph.type = arIn[0];
            icmph.code = arIn[1];
            idx += 2;

            arTemp = new byte[sizeof(ushort)]; //chksum is ushort
            Array.Copy(arIn, idx, arTemp, 0, arTemp.Length);
            idx += arTemp.Length;
            icmph.chksum = BitConverter.ToUInt16(arTemp, 0);

            //arTemp = new byte[sizeof(ushort)]; //id is the same length with chksum
            Array.Copy(arIn, idx, arTemp, 0, arTemp.Length);
            idx += arTemp.Length;
            icmph.id = BitConverter.ToUInt16(arTemp, 0);

            //arTemp = new byte[sizeof(ushort)]; //seq is the same length with id
            Array.Copy(arIn, idx, arTemp, 0, arTemp.Length);
            idx += arTemp.Length;
            icmph.seq = BitConverter.ToUInt16(arTemp, 0);

            arTemp = new byte[sizeof(ulong)]; //timestamp is ulong
            Array.Copy(arIn, idx, arTemp, 0, arTemp.Length);
            idx += arTemp.Length;
            icmph.timestamp = BitConverter.ToUInt64(arTemp, 0);
            
            return icmph;
        }
    }

    interface IMassPing
    {
        bool ping(string strHostName);  // 1 = success, 0 = failure
        int Frequency { set; }
        int Timeout { set;}             //default = 500msec
        int AvgTime { get;}             //measured average response time
        int AvgTTL { get;}             //measured average number of routing nodes the ping traveled
        string ErrorMessage { get;}        //a verbose error message in case of failure 
    }

    public class CMassPing : IMassPing
    {
        // create an event delegant
        public delegate void PingReadyEventHandler(object sender, PingEventArgs e);
        // create an event to notify subscribers
        public event PingReadyEventHandler PingReadyEvent;

        private Socket mSock = null;
        private System.Threading.Thread recvingThread = null;
        private System.Timers.Timer sendingTimer = null;
        private bool mbStarted = false;
        private IPEndPoint mRemoteEP;
        private ICMPHeader mIcmpHeader;
        
        private int m_Timeout;   //in msec
        private int m_Freq;      //in msec
        private int m_arTime;    //response times statistic
        private byte m_arTTL;     //routing stations statistic
        private string m_strErrorMessage;
        private bool m_arResults; //results of pings

        //----------------------------------------------------

        public string RemoteAddress
        {
            set
            {
                string strHostName = value;
                const int echoPort = 7;
                if (true == Char.IsDigit(strHostName[0]))
                {
                    IPAddress ipAddr = IPAddress.Parse(strHostName);
                    mRemoteEP = new IPEndPoint(ipAddr, echoPort);
                }
                else
                {
                    IPHostEntry lipa = Dns.GetHostEntry(strHostName);
                    mRemoteEP = new IPEndPoint(lipa.AddressList[0], echoPort);
                }
            }
        }
        public int Timeout { set { m_Timeout = Math.Max(value, 1); } }
        public int Frequency { set { m_Freq = Math.Max(value, 1); } }
        public int AvgTime { get { return m_arTime; } }
        public int AvgTTL { get { return m_arTTL; } }
        public string ErrorMessage { get { return m_strErrorMessage; } }

        public CMassPing()
        {
            m_Timeout = 1000; //msec
            m_Freq = 500; //mse
            m_strErrorMessage = "Don't know what happened";

            //set the ICMP header
            mIcmpHeader = new ICMPHeader();
            mIcmpHeader.type = 8; //ICMP_ECHO type
            mIcmpHeader.code = 0;
            mIcmpHeader.id = Convert.ToUInt16(Process.GetCurrentProcess().Id);  //lossy conversion
            mIcmpHeader.seq = 0;


            //create an ICMP socket
            mSock = new Socket(AddressFamily.InterNetwork, SocketType.Raw, ProtocolType.Icmp);
            mSock.Bind(new IPEndPoint(IPAddress.Any, 0)); //icmp don't necessary to be bind. but it is required sometimes
            
            sendingTimer = new System.Timers.Timer();
            sendingTimer.Elapsed += new ElapsedEventHandler(sendingTimer_Elapsed);
            recvingThread = new Thread(new ThreadStart(recvfun));
            mbStarted = false;
        }

        public bool Start()
        {
            if (mbStarted) return true;

            if (recvingThread.ThreadState == System.Threading.ThreadState.Unstarted)
                recvingThread.Start();
            else if (recvingThread.ThreadState == System.Threading.ThreadState.Aborted || recvingThread.ThreadState == System.Threading.ThreadState.Stopped)
            {
                recvingThread = new Thread(new System.Threading.ThreadStart(recvfun));
                recvingThread.Start();
            }

            sendingTimer.Interval = m_Freq;
            sendingTimer.Start();

            mbStarted = true;
            return true;
        }

        public bool Stop()
        {
            if (!mbStarted) return true;

            sendingTimer.Stop();
            recvingThread.Abort();

            mbStarted = false;
            return true;
        }

        public bool Dispose()
        {
            if (Stop())
            {
                sendingTimer.Close();
                mSock.Close();
            }
            return true;
        }

        private void sendingTimer_Elapsed(object sender, ElapsedEventArgs e)
        {
            //set up ICMPHeader structure
            mIcmpHeader.seq++;
            mIcmpHeader.chksum = 0;
            mIcmpHeader.timestamp = Convert.ToUInt64(DateTime.Now.Ticks);

            //fill header into byte array
            byte[] arHeader = mIcmpHeader.toByteArray();

            //create the array which is to be sent
            byte[] arBytes = new byte[arHeader.Length + 32];

            //copy header to array which is to be sent
            arHeader.CopyTo(arBytes, 0);

            mIcmpHeader.chksum = this.generateIPChecksum(arBytes);
            //now this is goofy because after we inserted the checksum into
            //the header we have to recreate the byte array
            arHeader = mIcmpHeader.toByteArray();
            arHeader.CopyTo(arBytes, 0);

            mSock.SendTo(arBytes, mRemoteEP);
        }

        private void recvfun()
        {
            while (true)
            {
                // check socket receive asnychronously and heed timeout
                if (true == mSock.Poll(1000 * m_Timeout, SelectMode.SelectRead))
                {

                    // Creates an IpEndPoint to capture the identity of the sending host.
                    IPEndPoint sender1 = new IPEndPoint(IPAddress.Any, 0);
                    EndPoint tempRemoteEP = (EndPoint)sender1;

                    // Creates a byte buffer to receive the message.
                    byte[] receiveBuffer = new byte[128]; //20(iph) + 22(icmph) + 32(data) = 74 < 128

                    // Receives datagram from a remote host.  This call blocks!
                    int nReceived = mSock.ReceiveFrom(receiveBuffer, ref tempRemoteEP);

                    ICMPHeader icmpHeader;
                    if (GetICMPHeaderFromRawData(receiveBuffer, nReceived, out icmpHeader))
                    {
                        m_arTTL = receiveBuffer[8]; // in the IP header

                        //calculate elapsed time         
                        DateTime now = DateTime.Now;
                        TimeSpan elapsedTime = new TimeSpan(now.Ticks - Convert.ToInt64(icmpHeader.timestamp));
                        m_arTime = elapsedTime.Milliseconds;
                       
                        PingReadyEvent(this, new PingEventArgs(true, m_arTime, m_arTTL, ""));
                    }
                    else
                        PingReadyEvent(this, new PingEventArgs(false, 0, 0, m_strErrorMessage));
                }
                else
                {
                    m_strErrorMessage = "Echoing socket is not readible in the given timeout";
                    PingReadyEvent(this, new PingEventArgs(false, 0, 0, m_strErrorMessage));
                }
            }
        }

        private bool GetICMPHeaderFromRawData(byte[] recvBuffer, int nLen, out ICMPHeader icmphdr)
        {
            icmphdr = new ICMPHeader();

            //the first 4 bits are the length of the IP Header in DWORDS
            int nIPHeaderLen = recvBuffer[0] & 0x0f;
            nIPHeaderLen *= 4; //in bytes

            //enough data received
            if (nLen < nIPHeaderLen + ICMPHeader.ICMPHEADERLENGTH)
            {
                m_strErrorMessage = "Not enough data echoed";
                return false;
            }

            byte[] arHeader = new byte[ICMPHeader.ICMPHEADERLENGTH]; // 22 is the ip header length
            Array.Copy(recvBuffer, nIPHeaderLen, arHeader, 0, ICMPHeader.ICMPHEADERLENGTH);

            icmphdr = ICMPHeader.fromByteArray(arHeader);

            //check if it is an ICMP_ECHOREPLY packet
            if (icmphdr.type != 0)
            {
                m_strErrorMessage = "Wrong type of telegram echoed";
                return false;
            }

            //check echoed process id is ours
            if (icmphdr.id != Convert.ToUInt16(Process.GetCurrentProcess().Id))  //lossy conversion
            {
                m_strErrorMessage = "Received echoed data was not sent by this process";
                return false;
            }

            return true;
        }

        public bool ping(string strHostName)
        {
            m_strErrorMessage = "No error occured";
            m_arResults = false;

            this.clearStats();

            //convert strHostName to an IPEndPoint
            try
            {
                IPEndPoint lep;

                //check if strHostname is already a dotted address
                //and create an IPEndpoint from it
                //note: port 7 is for echo but is irrelevant because of the socket type we are going to use
                const int echoPort = 7;

                if (true == Char.IsDigit(strHostName[0]))
                {
                    IPAddress ipAddr = IPAddress.Parse(strHostName);
                    lep = new IPEndPoint(ipAddr, echoPort);
                }
                else
                {
                    IPHostEntry lipa = Dns.GetHostEntry(strHostName);
                    lep = new IPEndPoint(lipa.AddressList[0], echoPort);
                }

                //number of tries
                if (true == tryPing(lep))
                {
                    m_arResults = true;
                    PingReadyEvent(this, new PingEventArgs(true, m_arTime, m_arTTL, m_strErrorMessage));
                }
                else
                {
                    PingReadyEvent(this, new PingEventArgs(false, m_arTime, m_arTTL, m_strErrorMessage));
                }
            }
            catch (SocketException ex)
            {
                
                m_strErrorMessage = ex.Message;
                PingReadyEvent(this, new PingEventArgs(false, m_arTime, m_arTTL, m_strErrorMessage));
            }
            catch (Exception ex)
            {
                m_strErrorMessage = ex.Message;
                PingReadyEvent(this, new PingEventArgs(false, m_arTime, m_arTTL, m_strErrorMessage));
            }

            return m_arResults;
        }

        private bool tryPing(IPEndPoint lep)
        {
            //do the ping
            bool bResult = false;

            //create an ICMP socket
            Socket sock = new Socket(AddressFamily.InterNetwork, SocketType.Raw, ProtocolType.Icmp);

            //set up ICMPHeader structure
            ICMPHeader header = new ICMPHeader();
            header.type = 8; //ICMP_ECHO type
            header.code = 0;
            header.id = Convert.ToUInt16(Process.GetCurrentProcess().Id);  //lossy conversion
            header.seq = 0;
            header.chksum = 0;
            DateTime startTime = DateTime.Now;
            header.timestamp = Convert.ToUInt64(startTime.Ticks);

            //fill header into byte array
            byte[] arHeader = header.toByteArray();

            //create the array which is to be sent
            byte[] arBytes = new byte[arHeader.Length + 32];

            //copy header to array which is to be sent
            arHeader.CopyTo(arBytes, 0);

            header.chksum = this.generateIPChecksum(arBytes);
            //now this is goofy because after we inserted the checksum into
            //the header we have to recreate the byte array
            arHeader = header.toByteArray();
            arHeader.CopyTo(arBytes, 0);

            sock.SendTo(arBytes, lep);

            // check socket receive asnychronously and heed timeout
            if (true == sock.Poll(1000 * m_Timeout, SelectMode.SelectRead))
            {

                // Creates an IpEndPoint to capture the identity of the sending host.
                IPEndPoint sender1 = new IPEndPoint(IPAddress.Any, 0);
                EndPoint tempRemoteEP = (EndPoint)sender1;

                // Creates a byte buffer to receive the message.
                byte[] receiveBuffer = new byte[128];

                // Receives datagram from a remote host.  This call blocks!
                int nReceived = sock.ReceiveFrom(receiveBuffer, ref tempRemoteEP);

                bResult = this.verifyReceivedMessage(receiveBuffer, nReceived, arHeader.Length);

                m_arTTL = receiveBuffer[8];
            }
            else
            {
                m_strErrorMessage = "Echoing socket is not readible in the given timeout";
            }

            //number of routing stations is part of the IP Header

            //calculate elapsed time         
            DateTime now = DateTime.Now;
            TimeSpan elapsedTime = new TimeSpan(now.Ticks - startTime.Ticks);
            m_arTime = elapsedTime.Milliseconds;

            return bResult;
        }

        private ushort generateIPChecksum(byte[] arBytes)
        {
            //generate an IP checksum based on a given data buffer
            ulong chksum = 0;

            int nSize = arBytes.Length;
            int i = 0;

            while (nSize > 1)
            {
                chksum += (ulong)((((ushort)arBytes[i + 1]) << 8) + (ushort)arBytes[i]);
                nSize--;
                nSize--;
                i++;
                i++;
            }

            if (nSize > 0)
            {
                chksum += arBytes[i];
            }

            chksum = (chksum >> 16) + (chksum & 0xffff);
            chksum += (chksum >> 16);

            ushort result = (ushort)(~chksum);
            return result;
        }

        private bool verifyReceivedMessage(byte[] arBytes, int nReceived, int minLengthSent)
        {
            //the first 4 bits are the length of the IP Header in DWORDS
            int nLengthIPHeader = arBytes[0] & 0x0f;
            nLengthIPHeader *= 4; //in bytes

            //enough data received
            if (nLengthIPHeader + minLengthSent > nReceived)
            {
                m_strErrorMessage = "Not enough data echoed";
                return false;
            }

            //check if it is an ICMP_ECHOREPLY packet
            if (arBytes[nLengthIPHeader] != 0)
            {
                m_strErrorMessage = "Wrong type of telegram echoed";
                return false;
            }

            int nId = arBytes[nLengthIPHeader + 6] + arBytes[nLengthIPHeader + 7] * 256;

            //check echoed process id is ours
            if (nId != Convert.ToUInt16(Process.GetCurrentProcess().Id))  //lossy conversion
            {
                m_strErrorMessage = "Received echoed data was not sent by this process";
                return false;
            }

            return true;
        }

        private void clearStats()
        {
            //clear statistical data
            m_arTime = 0;
            m_arTTL = 0;
        }
    }
}

public class PingEventArgs : EventArgs
{
    public bool m_bResult
    {
        get { return mResult; }
    } private bool mResult;

    public int m_nTime
    {
        get { return mTime; } 

    } private int mTime;

    public int m_nTTL 
    {
        get { return mTTL; }
    } private int mTTL;
    
    public string m_strErr
    {
        get { return mErr; }
    } private string mErr;

    public PingEventArgs(bool r, int t, int l, string e)
    {
        mResult = r;
        mTime = t;
        mTTL = l;
        mErr = e;
    }
}
